/*
 * $Id: JavadoXML.java,v 1.21 2003/04/26 15:20:55 znerd Exp $
 */
package org.znerd.javadoxml;

import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.ConstructorDoc;
import com.sun.javadoc.Doc;
import com.sun.javadoc.Doclet;
import com.sun.javadoc.ExecutableMemberDoc;
import com.sun.javadoc.FieldDoc;
import com.sun.javadoc.MethodDoc;
import com.sun.javadoc.PackageDoc;
import com.sun.javadoc.Parameter;
import com.sun.javadoc.ProgramElementDoc;
import com.sun.javadoc.RootDoc;
import com.sun.javadoc.Tag;
import java.io.IOException;
import java.io.File;
import java.io.FileWriter;
import java.io.Writer;
import java.io.UnsupportedEncodingException;
import org.znerd.xmlenc.XMLOutputter;

/**
 * Doclet that generates XML output.
 *
 * <p>This doclet recognizes the following options:
 *
 * <dl>
 *    <dt><code>-out</code></dt>
 *    <dd>The output file to write to. The default is <code>apidocs.xml</code>.</dd>
 *    <dt><code>-indent</code></dt>
 *    <dd>The indentation to use, if any. The default is no indentation..</dd>
 *    <dt><code>-newlines</code></dt>
 *    <dd>Flag that indicates if newlines should be used. Valid values are
 *    <code>true</code> or <code>yes</code> if newlines should be used, and
 *    <code>false</code> or <code>no</code> if no newlines should be used. The
 *    default is <code>no</code>.</dd>
 *    <dt><code>-encoding</code></dt>
 *    <dd>The encoding to use. The default is <code>UTF-8</code>.</dd>
 * </dl>
 *
 * @version $Revision: 1.21 $ $Date: 2003/04/26 15:20:55 $
 * @author Ernst de Haan (<a href="mailto:znerd@FreeBSD.org">znerd@FreeBSD.org</a>)
 */
public final class JavadoXML extends Doclet {

   //-------------------------------------------------------------------------
   // Class fields
   //-------------------------------------------------------------------------

   /**
    * The default output filename.
    */
   private static final String DEFAULT_FILENAME = "apidocs.xml";

   /**
    * The default destination directory. The value of this constant is
    * <code>"."</code>, the current directory.
    */
   private static final String DEFAULT_DESTDIR = ".";

   /**
    * The default indentation. The value of this constant is <code>""</code>,
    * the empty string.
    */
   private static final String DEFAULT_INDENT = "";

   /**
    * The default setting for newlines. By default no newlines will be
    * inserted.
    */
   private static final boolean DEFAULT_NEWLINES = false;

   /**
    * The default encoding. The value of this constant is
    * <code>"UTF-8"</code>.
    */
   private static final String DEFAULT_ENCODING = "UTF-8";


   //-------------------------------------------------------------------------
   // Class functions
   //-------------------------------------------------------------------------

   /**
    * Checks a doclet option.
    *
    * @param option
    *    the doclet option to check, not <code>null</code>.
    *
    * @return
    *    the number of arguments the option has, or <code>0</code> (zero) if
    *    the argument is unknown, or a negative value in case of an error.
    *
    * @throws IllegalArgumentException
    *    if <code>option == null</code>.
    */
   public static int optionLength(String option)
   throws IllegalArgumentException {

      // Check preconditions
      if (option == null) {
         throw new IllegalArgumentException("option == null");
      }

      if ("-out".equals(option)
       || "-d".equals(option)
       || "-indent".equals(option)
       || "-newlines".equals(option)
       || "-encoding".equals(option)) {
         return 2;
      } else {
         return 0;
      }
   }

   /**
    * Generates the XML API documentation.
    *
    * @param root
    *    the packages and/or classes to document, not <code>null</code>.
    *
    * @return
    *    <code>true</code> on success, <code>false</code> otherwise.
    *
    * @throws IllegalArgumentException
    *    if <code>root == null</code>.
    */
   public static boolean start(RootDoc root) throws IllegalArgumentException {

      // Check preconditions
      if (root == null) {
         throw new IllegalArgumentException("root == null");
      }

      // Runs the actual doclet
      JavadoXML doclet = new JavadoXML();
      return doclet.run(root);
   }


   //-------------------------------------------------------------------------
   // Constructor
   //-------------------------------------------------------------------------

   /**
    * Constructs a new <code>JavadoXML</code> object.
    */
   private JavadoXML() {
      // empty
   }


   //-------------------------------------------------------------------------
   // Fields
   //-------------------------------------------------------------------------

   /**
    * Flag that indicates if this doclet is currently running. One instance
    * can only run within one thread at the same time.
    */
   private transient boolean _running;

   /**
    * The starting point for this Javadoc run. This field is initialized by
    * {@link #run(RootDoc)}.
    */
   private transient RootDoc _root;

   /**
    * The name of the file to write the output to. This is a path relative to
    * the {@link #_destdir destination directory}.
    *
    * <p>The default is {@link #DEFAULT_FILENAME}.</p>
    */
   private String _filename = DEFAULT_FILENAME;

   /**
    * The parent directory for the output file.
    *
    * <p>The default is {@link #DEFAULT_DESTDIR}.</p>
    */
   private String _destdir = DEFAULT_DESTDIR;

   /**
    * The indentation to use.
    *
    * <p>The default is {@link #DEFAULT_INDENT}.</p>
    */
   private String _indent = DEFAULT_INDENT;

   /**
    * Flag that indicates if newlines should be output.
    *
    * <p>The default is {@link #DEFAULT_NEWLINES}.</p>
    */
   private boolean _newlines = DEFAULT_NEWLINES;

   /**
    * The encoding to use for the output file.
    *
    * <p>The default is {@link #DEFAULT_ENCODING}.</p>
    */
   private String _encoding = DEFAULT_ENCODING;

   /**
    * The <code>XMLOutputter</code> that is used to generate XML with.
    */
   private XMLOutputter _outputter;


   //-------------------------------------------------------------------------
   // Methods
   //-------------------------------------------------------------------------

   /**
    * Runs this doclet.
    *
    * @param root
    *    the packages and/or classes to document, not <code>null</code>.
    *
    * @return
    *    <code>true</code> on success, <code>false</code> otherwise.
    *
    * @throws IllegalArgumentException
    *    if <code>root == null</code>.
    *
    * @throws IllegalStateException
    *    if this doclet is currently already running.
    */
   public boolean run(RootDoc root)
   throws IllegalArgumentException, IllegalStateException {

      // Check preconditions
      if (root == null) {
         throw new IllegalArgumentException("root == null");
      }
      synchronized (JavadoXML.class) {
         if (_running == true) {
            throw new IllegalStateException("This doclet is currently already running.");
         }
         _running = true;
      }

      _root = root;
      return runImpl();
   }

   private boolean runImpl() {

      // Interpret the options
      interpretOptions();

      // Create a File object for the destdir
      File dir = new File(_destdir);
      if (dir.exists() == false) {
         _root.printError("Destination directory \"" + _destdir + "\" does not exist.");
      } else if (dir.isDirectory() == false) {
         _root.printError("Destination directory \"" + _destdir + "\" is not actually a directory.");
      }

      // See if we can open the output file for writing
      File out = new File(dir, _filename);
      String outAbs = out.getAbsolutePath();
      if (out.exists()) {
         if (out.canWrite()) {
            // log("Output file \"" + outAbs + "\" exists and is writable.", Project.MSG_DEBUG);
         } else {
            _root.printError("Output file \"" + outAbs + "\" exists and is not writable.");
            return false;
         }
      } else if (out.isDirectory()) {
         _root.printError("The file specified in the \"out\" property (\"" + outAbs + "\") is a directory.");
         return false;
      } else {
         try {
            out.createNewFile();
         } catch (IOException ioException) {
            _root.printError("Unable to create output file \"" + outAbs + "\": " + ioException.getMessage());
            return false;
         } catch (SecurityException securityException) {
            _root.printError("Not allowed by security manager to create output file \"" + outAbs + "\": " + securityException.getMessage());
            return false;
         }
      }

      // Create a stream to the output file
      Writer writer;
      try {
         writer = new FileWriter(out);
      } catch (IOException exception) {
         _root.printError("Unable to create a stream to the output file \"" + outAbs + "\": " + exception.getMessage());
         return false;
      }

      // Create the XML document
      // TODO: _indent and _newlines are ignored
      try {
         _outputter = new XMLOutputter(writer, _encoding);
      } catch (UnsupportedEncodingException uee) {
         _root.printError("The encoding \"" + _encoding + "\" is not supported.");
         return false;
      }

      // Output the XML
      try {
         createDocument();
         _outputter.close();
         writer.flush();
         writer.close();
      } catch (IOException exception) {
         _root.printError("Failed to write XML document to output file \"" + outAbs + "\": " + exception.getMessage());
         return false;
      }

      return true;
   }

   /**
    * Interprets the options.
    */
   private void interpretOptions() {

      // Get the options
      String[][] options = _root.options();
      int optionCount = (options==null) ? 0 : options.length;

      // Get all overrides
      for (int i = 0; i < optionCount; i++) {
         String name = options[i][0];

         // File name
         if ("-out".equals(name)) {
            _filename = options[i][1];

         // Destination directory
         } else if ("-d".equals(name)) {
            _destdir = options[i][1];

         // Indentation
         } else if ("-indent".equals(name)) {
            _indent = options[i][1];

         // Newlines
         } else if ("-newlines".equals(name)) {
            String value = options[i][1];
            value = value.trim().toLowerCase();
            if ("true".equals(value) || "yes".equals(value)) {
               _newlines = true;
            } else if ("false".equals(value) || "no".equals(value)) {
               _newlines = false;
            } else {
               _root.printWarning("The option -newlines has an illegal value: " + options[i][1]);
            }

         // Encoding
         } else if ("-encoding".equals(name)) {
            _encoding = options[i][1];
         }
      }
   }

   /**
    * Writes the XML document.
    *
    * @throws IOException
    *    if there was an I/O error.
    */
   private void createDocument()
   throws IOException {

      // Open the root element
      _outputter.startTag("javadoc");

      // Handle all specified packages and classes
      handle(_root.specifiedPackages());
      handle(_root.specifiedClasses());

      // Close the root element
      _outputter.endTag();
   }

   /**
    * Creates XML output for the specified packages.
    *
    * <p>This method delegates to {@link #handle(PackageDoc)}.</p>
    *
    * @param packages
    *    the packages to be processed, can be <code>null</code>.
    *
    * @throws IOException
    *    if there was an I/O error.
    */
   private void handle(PackageDoc[] packages)
   throws IOException {
      int count = (packages == null) ? 0 : packages.length;
      for (int i = 0; i < count; i++) {
         PackageDoc p = packages[i];
         handle(p);
      }
   }

   /**
    * Creates XML output for an individual package.
    *
    * @param packages
    *    the package to be processed, not <code>null</code>.
    */
   private void handle(PackageDoc p)
   throws IOException {

      _outputter.startTag("package");
      _outputter.attribute("name", p.name());
      handleDoc(p);

      ClassDoc[] classes = p.allClasses();
      int count = (classes == null) ? 0 : classes.length;
      for (int i = 0; i < count; i++) {
         ClassDoc c = classes[i];
         handle(c);
      }

      _outputter.endTag();
   }

   private void handleDoc(Doc doc)
   throws IOException {
      Tag[] tags = doc.inlineTags();
      int count = (tags == null) ? 0 : tags.length;
      if (count > 0) {
         _outputter.startTag("comment");
         for (int i = 0; i < count; i++) {
            handle(tags[i]);
         }
         _outputter.endTag();
      }
   }

   private void handle(Tag tag)
   throws IOException {
      String kind = tag.kind().trim().toLowerCase();
      if ("text".equals(kind)) {
         _outputter.pcdata(tag.text());
      } else {
         String name = tag.name().trim().toLowerCase();
         if (kind.charAt(0) == '@') {
            kind = kind.substring(1);
         }
         if (name.charAt(0) == '@') {
            name = name.substring(1);
         }

         _outputter.startTag("tag");
         _outputter.attribute("kind", kind);
         _outputter.attribute("name", name);
         _outputter.pcdata(tag.text());
         _outputter.endTag();
      }
   }

   /**
    * Creates XML output for the specified classes.
    *
    * <p>This method delegates to
    * {@link #handle(ClassDoc)}.</p>
    *
    * @param classes
    *    the classes to be processed, can be <code>null</code>.
    */
   private void handle(ClassDoc[] classes)
   throws IOException {

      int count = (classes == null) ? 0 : classes.length;
      for (int i = 0; i < count; i++) {
         ClassDoc c = classes[i];
         handle(c);
      }
   }

   private void handle(ClassDoc c)
   throws IOException {

      // Determine the element name
      // XXX: Use ?: operator here
      String elementName;
      if (c.isInterface()) {
         elementName = "interface";
      } else {
         elementName = "class";
      }

      _outputter.startTag(elementName);

      // Set the name of the class/interface
      _outputter.attribute("name", c.name());

      // Some attributes that do not apply to interfaces
      if (c.isInterface() == false) {

         // Set the 'abstract' flag
         attribute("abstract",  c.isAbstract());
         attribute("final",     c.isFinal());
         attribute("static",    c.isStatic());
         attribute("exception", c.isException());
         attribute("error",     c.isError());
      }
      handleProgramElement(c);
      handleDoc(c);

      handle(c.constructors());
      handle(c.methods());
      handle(c.fields());
      handle(c.innerClasses());

      _outputter.endTag();
   }

   /**
    * Creates XML output for the specified constructors.
    *
    * <p>This method delegates to
    * {@link #handle(ConstructorDoc)}.</p>
    *
    * @param constructors
    *    the constructors to be processed, can be <code>null</code>.
    */
   private void handle(ConstructorDoc[] constructors)
   throws IOException {

      int count = (constructors == null) ? 0 : constructors.length;
      for (int i = 0; i < count; i++) {
         ConstructorDoc c = constructors[i];
         handle(c);
      }
   }

   private void handle(ConstructorDoc c)
   throws IOException {

      // Determine the element name
      _outputter.startTag("constructor");
      handleProgramElement(c);
      handleDoc(c);
      handleParameters(c);
      _outputter.endTag();
   }

   /**
    * Creates XML output for the specified methods.
    *
    * <p>This method delegates to
    * {@link #handle(MethodDoc)}.</p>
    *
    * @param methods
    *    the methods to be processed, can be <code>null</code>.
    */
   private void handle(MethodDoc[] methods)
   throws IOException {

      int count = (methods == null) ? 0 : methods.length;
      for (int i = 0; i < count; i++) {
         MethodDoc m = methods[i];
         handle(m);
      }
   }

   private void handle(MethodDoc m)
   throws IOException {

      // Determine the element name
      _outputter.startTag("method");

      // Set the name of the class/interface
      _outputter.attribute("name", m.name());

      // Set the 'abstract' flag
      attribute("abstract", m.isAbstract());
      attribute("final",    m.isFinal());
      attribute("static",   m.isStatic());
      _outputter.attribute("return", m.returnType().toString());
      handleProgramElement(m);
      handleDoc(m);
      handleParameters(m);
      _outputter.endTag();
   }

   /**
    * Creates XML output for the specified fields.
    *
    * <p>This method delegates to
    * {@link #handle(MethodDoc)}.</p>
    *
    * @param fields
    *    the fields to be processed, can be <code>null</code>.
    */
   private void handle(FieldDoc[] fields)
   throws IOException {

      int count = (fields == null) ? 0 : fields.length;
      for (int i = 0; i < count; i++) {
         FieldDoc f = fields[i];
         handle(f);
      }
   }

   private void handle(FieldDoc f)
   throws IOException {

      // Determine the element name
      _outputter.startTag("field");

      // Set the name of the class/interface
      _outputter.attribute("name", f.name());

      // Set the 'abstract' flag
      attribute("final",     f.isFinal());
      attribute("static",    f.isStatic());
      attribute("transient", f.isTransient());
      attribute("volatile",  f.isVolatile());
      _outputter.attribute("type", f.type().toString());

      handleDoc(f);

      _outputter.endTag();
   }

   private void handleProgramElement(ProgramElementDoc p)
   throws IOException {
      String access;
      if (p.isPrivate()) {
         access = "private";
      } else if (p.isPackagePrivate()) {
         access = "package";
      } else if (p.isProtected()) {
         access = "protected";
      } else if (p.isPublic()) {
         access = "public";
      } else {
         throw new InternalError("The specified program element is neither private nor package-private nor protected nor public.");
      }
      _outputter.attribute("access", access);
   }

   private void handleParameters(ExecutableMemberDoc d)
   throws IOException {
      Parameter[] parameters = d.parameters();
      int count = (parameters == null) ? 0 : parameters.length;
      for (int i = 0; i < count; i++) {
         Parameter p = parameters[i];
         handle(p);
      }
   }

   private void handle(Parameter p)
   throws IOException {
      _outputter.startTag("parameter");
      _outputter.attribute("name", p.name());
      _outputter.attribute("type", p.type().toString());
      _outputter.endTag();
   }

   private void attribute(String attribute, boolean cond)
   throws IOException {
      String value = cond ? "true" : "false";
      _outputter.attribute(attribute, value);
   }
}
